@@ -1,3 +1,5 @@
+use std::collections::HashMap;
+
 use aws_sdk_ec2 as ec2;
 use tokio_stream::StreamExt;
 
@@ -10,6 +12,8 @@ mod impls;
 #[derive(Debug)]
 pub(crate) struct Ec2Resources {
     client: ec2::Client,
+    tags: Vec<(String, String)>,
+    tag_descriptions: Vec<ec2::model::TagDescription>,
     vpcs: Vec<ec2::model::Vpc>,
     subnets: Vec<ec2::model::Subnet>,                      // 1
     instances: Vec<ec2::model::Instance>,                  // 2
@@ -26,9 +30,12 @@ pub(crate) struct Ec2Resources {
 }
 
 impl Ec2Resources {
-    pub(crate) fn new(client: ec2::Client) -> Self {
+    pub(crate) fn new(client: ec2::Client, tags: &[(String, String)]) -> Self {
+        let tags = tags.to_vec();
         Self {
             client,
+            tags,
+            tag_descriptions: vec![],
             vpcs: vec![],
             subnets: vec![],
             instances: vec![],
@@ -75,7 +82,49 @@ impl Ec2Resources {
     }
 
     pub(crate) fn trees(&self) -> impl Iterator<Item = ptree::item::StringItem> + '_ {
-        self.vpcs().iter().map(|vpc| self.vpc_tree(vpc))
+        if self.tag_descriptions.is_empty() {
+            self.vpcs()
+                .iter()
+                .map(|vpc| self.vpc_tree(vpc))
+                .collect::<Vec<_>>()
+                .into_iter()
+        } else {
+            vec![self.tag_tree()].into_iter()
+        }
+    }
+
+    fn tag_tree(&self) -> ptree::item::StringItem {
+        let mut tags: HashMap<&str, HashMap<&str, HashMap<&str, Vec<&str>>>> = HashMap::new();
+
+        for tag in self.tag_descriptions.iter() {
+            tags.entry(tag.key().unwrap_or_default())
+                .or_default()
+                .entry(tag.value().unwrap_or_default())
+                .or_default()
+                .entry(tag.resource_type().map(|r| r.as_str()).unwrap_or_default())
+                .or_default()
+                .push(tag.resource_id().unwrap_or_default());
+        }
+
+        let mut tree = ptree::TreeBuilder::new(String::from("Tags"));
+        let tree = &mut tree;
+
+        for (tag, values) in tags {
+            tree.begin_child(tag.to_string());
+            for (value, resources) in values {
+                tree.begin_child(value.to_string());
+                for (resource_type, resource_ids) in resources {
+                    tree.begin_child(resource_type.to_string());
+                    for resource_id in resource_ids {
+                        tree.add_empty_child(resource_id.to_string());
+                    }
+                    tree.end_child();
+                }
+                tree.end_child();
+            }
+            tree.end_child();
+        }
+        tree.build()
     }
 
     fn vpc_tree(&self, vpc: &ec2::model::Vpc) -> ptree::item::StringItem {
@@ -207,6 +256,30 @@ impl Ec2Resources {
             .collect()
     }
 
+    pub(crate) async fn collect_tags(
+        &mut self,
+        _progress: &indicatif::ProgressBar,
+    ) -> Result<(), ec2::Error> {
+        self.tag_descriptions = self
+            .client
+            .describe_tags()
+            .into_paginator()
+            .items()
+            .send()
+            // .map(|tag_description| {
+            //     if let Ok(ref tag_description) = tag_description {
+            //         if let Some(key) = tag_description.key() {
+            //             progress.set_message(key.to_string())
+            //         }
+            //         progress.inc(1);
+            //     }
+            //     tag_description
+            // })
+            .collect::<Result<_, _>>()
+            .await?;
+        Ok(())
+    }
+
     pub(crate) async fn collect_vpcs(&mut self, vpcs: &[String]) -> Result<(), ec2::Error> {
         self.vpcs = vpcs
             .iter()
@@ -217,6 +290,7 @@ impl Ec2Resources {
             .client
             .describe_vpcs()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -231,6 +305,7 @@ impl Ec2Resources {
             .client
             .describe_subnets()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -245,6 +320,7 @@ impl Ec2Resources {
             .client
             .describe_instances()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -262,6 +338,7 @@ impl Ec2Resources {
             .client
             .describe_internet_gateways()
             .optionally_filter(self.attachment_vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -276,6 +353,7 @@ impl Ec2Resources {
             .client
             .describe_route_tables()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -290,6 +368,7 @@ impl Ec2Resources {
             .client
             .describe_network_acls()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -304,6 +383,7 @@ impl Ec2Resources {
             .client
             .describe_vpc_peering_connections()
             .optionally_filter(self.requester_vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -318,6 +398,7 @@ impl Ec2Resources {
             .client
             .describe_vpc_endpoints()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -332,6 +413,7 @@ impl Ec2Resources {
             .client
             .describe_nat_gateways()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -346,6 +428,7 @@ impl Ec2Resources {
             .client
             .describe_security_groups()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -360,6 +443,7 @@ impl Ec2Resources {
             .client
             .describe_vpn_connections()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .send()
             .await?
             .vpn_connections
@@ -373,6 +457,7 @@ impl Ec2Resources {
             .client
             .describe_vpn_gateways()
             .optionally_filter(self.attachment_vpc_filter())
+            .fold_filters(self.tag_filter())
             .send()
             .await?
             .vpn_gateways
@@ -386,6 +471,7 @@ impl Ec2Resources {
             .client
             .describe_network_interfaces()
             .optionally_filter(self.vpc_filter())
+            .fold_filters(self.tag_filter())
             .into_paginator()
             .items()
             .send()
@@ -408,6 +494,14 @@ impl Ec2Resources {
         }
     }
 
+    fn tag_filter(&self) -> Vec<ec2::model::Filter> {
+        self.tags
+            .iter()
+            .map(|(key, value)| (format!("tag:{key}"), [value]))
+            .map(|(key, values)| filter(key, values))
+            .collect()
+    }
+
     fn attachment_vpc_filter(&self) -> Option<ec2::model::Filter> {
         let vpcs = self
             .vpcs
@@ -446,7 +540,10 @@ pub(crate) async fn regions(client: &ec2::Client) -> Result<Vec<ec2::model::Regi
     Ok(regions)
 }
 
-fn filter(key: &str, values: impl IntoIterator<Item = impl Into<String>>) -> ec2::model::Filter {
+fn filter(
+    key: impl Into<String>,
+    values: impl IntoIterator<Item = impl Into<String>>,
+) -> ec2::model::Filter {
     let builder = ec2::model::Filter::builder().name(key);
     values
         .into_iter()
